import binascii
from datetime import datetime
from datetime import timedelta

from ..errors import Error
from ..errors import EncodeError as _EncodeError
from ..errors import DecodeError as _DecodeError
from .. import compat


class EncodeError(_EncodeError):
    """General ASN.1 encode error.

    """

    pass


class DecodeError(_DecodeError):
    """General ASN.1 decode error with error location in the message.

    """

    def __init__(self, message):
        super(DecodeError, self).__init__()
        self.message = message
        self.location = []

    def __str__(self):
        return "{}: {}".format(': '.join(self.location[::-1]),
                               self.message)


class DecodeTagError(DecodeError):
    """ASN.1 tag decode error.

    """

    def __init__(self, type_name, expected_tag, actual_tag, offset):
        message = "Expected {} with tag '{}' at offset {}, but got '{}'.".format(
            type_name,
            binascii.hexlify(expected_tag).decode('ascii'),
            offset,
            binascii.hexlify(actual_tag).decode('ascii'))
        super(DecodeTagError, self).__init__(message)


class DecodeContentsLengthError(DecodeError):
    """ASN.1 contents length decode error.

    """

    def __init__(self, length, offset, contents_max):
        message = ('Expected at least {} contents byte(s) at offset {}, '
                   'but got {}.').format(length,
                                         offset,
                                         contents_max - offset)
        super(DecodeContentsLengthError, self).__init__(message)

        self.length = length
        self.offset = offset
        self.contents_max = contents_max


def format_or(items):
    """Return a string of comma separated items, with the last to items
    separated by "or".

    [1, 2, 3] -> "1, 2 or 3"

    """

    formatted_items = []

    for item in items:
        try:
            item = "'" + item + "'"
        except TypeError:
            item = str(item)

        formatted_items.append(item)

    if len(items) == 1:
        return formatted_items[0]
    else:
        return '{} or {}'.format(', '.join(formatted_items[:-1]),
                                 formatted_items[-1])


def utc_time_to_datetime(string):
    """Convert given ASN.1 UTC time string `string` to a
    ``datetime.datetime`` object.

    """

    length = len(string)

    if string[-1] == 'Z':
        if length == 11:
            return datetime.strptime(string[:10], '%y%m%d%H%M')
        elif length == 13:
            return datetime.strptime(string[:12], '%y%m%d%H%M%S')
        else:
            raise Error(
                "Expected an UTC time string, but got '{}'.".format(string))
    elif length == 15:
        return compat.strptime(string, '%y%m%d%H%M%z')
    elif length == 17:
        return compat.strptime(string, '%y%m%d%H%M%S%z')
    else:
        raise Error(
            "Expected an UTC time string, but got '{}'.".format(string))


def utc_time_from_datetime(date):
    """Convert given ``datetime.datetime`` object `date` to an ASN.1 UTC
    time string.

    """

    fmt = '%y%m%d%H%M'

    if date.second > 0:
        fmt += '%S'

    if date.tzinfo is None:
        fmt += 'Z'
    else:
        fmt += '%z'

    return date.strftime(fmt)


def restricted_utc_time_to_datetime(string):
    """Convert given restricted ASN.1 UTC time string `string` to a
    ``datetime.datetime`` object.

    """

    length = len(string)

    if string[-1] != 'Z':
        raise Error(
            "Expected a restricted UTC time string ending with 'Z', "
            "but got '{}'.".format(string))

    if length != 13:
        raise Error(
            "Expected a restricted UTC time string of length 13, "
            "but got '{}'.".format(string))

    return datetime.strptime(string[:-1], '%y%m%d%H%M%S')


def restricted_utc_time_from_datetime(date):
    """Convert given ``datetime.datetime`` object `date` to an restricted
    ASN.1 UTC time string.

    """

    if date.tzinfo is not None:
        date -= date.utcoffset()

    return date.strftime('%y%m%d%H%M%S') + 'Z'


def generalized_time_to_datetime(string):
    """Convert given ASN.1 generalized time string `string` to a
    ``datetime.datetime`` object.

    """

    length = len(string)

    if string[-1] == 'Z':
        if length == 15:
            date = datetime.strptime(string[:14], '%Y%m%d%H%M%S')
        else:
            date = datetime.strptime(string[:-1], '%Y%m%d%H%M%S.%f')

        return date.replace(tzinfo=compat.timezone(timedelta(hours=0)))
    elif string[-5] in '-+':
        if length == 19:
            return compat.strptime(string, '%Y%m%d%H%M%S%z')
        else:
            return compat.strptime(string, '%Y%m%d%H%M%S.%f%z')
    else:
        if length == 14:
            return compat.strptime(string, '%Y%m%d%H%M%S')
        else:
            return compat.strptime(string, '%Y%m%d%H%M%S.%f')


def generalized_time_from_datetime(date):
    """Convert given ``datetime.datetime`` object `date` to an ASN.1
    generalized time string.

    """

    if date.microsecond > 0:
        string = date.strftime('%Y%m%d%H%M%S.%f').rstrip('0')
    else:
        string = date.strftime('%Y%m%d%H%M%S')

    if date.tzinfo is not None:
        if date.utcoffset():
            string += date.strftime('%z')
        else:
            string += 'Z'

    return string


def restricted_generalized_time_to_datetime(string):
    """Convert given restricted ASN.1 generalized time string `string` to
    a ``datetime.datetime`` object.

    """

    if string[-1] != 'Z':
        raise Error(
            "Expected a restricted generalized time string ending with 'Z', "
            "but got '{}'.".format(string))

    if '.' in string:
        if string[-2] == '0':
            raise Error(
                "Expected a restricted generalized time string with no "
                "trailing zeros, but got '{}'.".format(string))

        return datetime.strptime(string[:-1], '%Y%m%d%H%M%S.%f')
    elif len(string) == 15:
        return datetime.strptime(string[:-1], '%Y%m%d%H%M%S')
    else:
        raise Error(
            "Expected a restricted generalized time string, but got '{}'.".format(
                string))


def restricted_generalized_time_from_datetime(date):
    """Convert given ``datetime.datetime`` object `date` to an restricted
    ASN.1 generalized time string.

    """

    if date.tzinfo is not None:
        date -= date.utcoffset()

    if date.microsecond > 0:
        string = date.strftime('%Y%m%d%H%M%S.%f').rstrip('0')
    else:
        string = date.strftime('%Y%m%d%H%M%S')

    return string + 'Z'



import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools


class Asn1ToolsCheckConstraintsTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_all_codecs(self):
        codecs = [
            'ber',
            'der',
            'gser',
            'jer',
            'per',
            'uper',
            'xer'
        ]

        for codec in codecs:
            foo = asn1tools.compile_string(
                "Foo DEFINITIONS AUTOMATIC TAGS ::= "
                "BEGIN "
                "A ::= INTEGER "
                "END",
                codec)

            with self.assertRaises(NotImplementedError):
                foo.check_constraints('A', 0)

    def test_integer(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= INTEGER "
            "B ::= INTEGER (5..99) "
            "C ::= INTEGER (-10..10) "
            "D ::= INTEGER (5..99, ...) "
            "E ::= INTEGER (1000..1000) "
            "F ::= SEQUENCE { "
            "  a INTEGER (4..4), "
            "  b INTEGER (40..40), "
            "  c INTEGER (400..400) "
            "} "
            "G ::= B (6..7) "
            "END")

        # Ok.
        datas = [
            ('A',  32768),
            ('A',      0),
            ('A', -32769),
            ('B',      5),
            ('B',      6),
            ('B',     99),
            ('C',    -10),
            ('C',     10),
            ('D',     99),
            ('E',   1000),
            ('F',   {'a': 4, 'b': 40, 'c': 400})
        ]

        for type_name, decoded in datas:
            with self.assertRaises(NotImplementedError):
                foo.check_constraints(type_name, decoded)

        # Not ok.
        datas = [
            ('B',      4, ': 4 does not fulfill 5..99'),
            ('B',    100, ': 100 does not fulfill 5..99'),
            ('C',    -11, ': -11 does not fulfill -10..10'),
            ('C',     11, ': 11 does not fulfill -10..10'),
            ('D',    100, ': 100 does not fulfill 5..99'),
            ('E',      0, ': 0 does not fulfill 1000..1000'),
            ('F',
             {'a': 4, 'b': 41, 'c': 400},
             'b: 41 does not fulfill 40..40')
        ]

        for type_name, decoded, message in datas:
            with self.assertRaises(NotImplementedError):
                with self.assertRaises(asn1tools.ConstraintsError) as cm:
                    foo.check_constraints(type_name, decoded)

                self.assertEqual(str(cm.exception), message)


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
from .utils import Asn1ToolsBaseTest
import asn1tools


CODECS = ['ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer']


class Asn1ToolsCodecsConsistencyTest(Asn1ToolsBaseTest):

    maxDiff = None

    def encode_decode_all_codecs(self, type_spec, values):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        foos = []

        for codec in CODECS:
            foos.append(asn1tools.compile_string(spec, codec))

        gser = asn1tools.compile_string(spec, 'gser')

        for value in values:
            decoded = value

            for foo in foos:
                encoded = foo.encode('A', decoded)
                decoded = foo.decode('A', encoded)
                self.assertEqual(type(decoded), type(value))
                self.assertEqual(decoded, value)

            gser.encode('A', decoded)

    def test_boolean(self):
        self.encode_decode_all_codecs("BOOLEAN", [True, False])

    def test_integer(self):
        self.encode_decode_all_codecs("INTEGER", [1, 123456789, -2, 0])

    def test_real(self):
        with self.assertRaises(NotImplementedError):
            self.encode_decode_all_codecs("REAL", [0.0, 1.0])

    def test_null(self):
        self.encode_decode_all_codecs("NULL", [None])

    def test_bit_string(self):
        self.encode_decode_all_codecs("BIT STRING",
                                      [(b'\x58', 5), (b'\x58\x80', 9)])

    def test_octet_string(self):
        self.encode_decode_all_codecs("OCTET STRING", [b'', b'\x12\x34'])

    def test_object_identifier(self):
        self.encode_decode_all_codecs("OBJECT IDENTIFIER", ['1.2.33'])

    def test_enumerated(self):
        self.encode_decode_all_codecs("ENUMERATED { a(0), b(5) }", ['a', 'b'])

    def test_sequence(self):
        self.encode_decode_all_codecs("SEQUENCE { a NULL }", [{'a': None}])

    def test_sequence_of(self):
        self.encode_decode_all_codecs("SEQUENCE OF NULL", [[], [None, None]])

    def test_set(self):
        self.encode_decode_all_codecs("SET { a NULL }", [{'a': None}])

    def test_set_of(self):
        self.encode_decode_all_codecs("SET OF NULL", [[], [None, None]])

    def test_choice(self):
        self.encode_decode_all_codecs("CHOICE { a NULL }", [('a', None)])

    def test_utf8_string(self):
        self.encode_decode_all_codecs("UTF8String", [u'hi'])

    def test_numeric_string(self):
        self.encode_decode_all_codecs("NumericString", [u'123'])

    def test_printable_string(self):
        self.encode_decode_all_codecs("PrintableString", [u'hi'])

    def test_ia5_string(self):
        self.encode_decode_all_codecs("IA5String", [u'hi'])

    def test_visible_string(self):
        self.encode_decode_all_codecs("VisibleString", [u'hi'])

    def test_general_string(self):
        self.encode_decode_all_codecs("GeneralString", [u'hi'])

    def test_bmp_string(self):
        self.encode_decode_all_codecs("BMPString", [u'hi'])

    def test_graphic_string(self):
        self.encode_decode_all_codecs("GraphicString", [u'hi'])

    def test_teletex_string(self):
        self.encode_decode_all_codecs("TeletexString", [u'hi'])

    def test_universal_string(self):
        self.encode_decode_all_codecs("UniversalString", [u'hi'])

    def test_utc_time(self):
        self.encode_decode_all_codecs("UTCTime", [datetime(2020, 3, 12)])

    def test_generalized_time(self):
        self.encode_decode_all_codecs("GeneralizedTime",
                                      [datetime(2021, 3, 12)])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from .utils import Asn1ToolsBaseTest
import sys
from copy import deepcopy
from asn1tools.codecs import utc_time_to_datetime as ut2dt
from asn1tools.codecs import generalized_time_to_datetime as gt2dt

import asn1tools

sys.path.append('tests/files')
sys.path.append('tests/files/3gpp')
sys.path.append('tests/files/ietf')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0
from rfc4511 import EXPECTED as RFC4511


class Asn1ToolsGserTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_real(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= REAL "
            "END",
            'gser')

        datas = [
            ('A',                     0.0, b'a A ::= 0'),
            ('A',            float('inf'), b'a A ::= PLUS-INFINITY'),
            ('A',           float('-inf'), b'a A ::= MINUS-INFINITY'),
            ('A',                     1.0, b'a A ::= 1.0E0'),
            ('A',                     1.1, b'a A ::= 1.1E0'),
            ('A',               1234.5678, b'a A ::= 1234.5678E0'),
            ('A',                       8, b'a A ::= 8E0'),
            ('A',                   0.625, b'a A ::= 0.625E0')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_null(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= NULL "
            "END",
            'gser')

        decoded = None
        encoded = b'a A ::= NULL'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_octet_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= OCTET STRING "
            "END",
            'gser')

        datas = [
            ('A',
             b'\x01\x23\x45\x67\x89\xab\xcd\xef',
             b"a A ::= '0123456789ABCDEF'H")
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_sequence(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a BOOLEAN, "
            "  b BOOLEAN OPTIONAL "
            "}"
            "B ::= SEQUENCE { "
            "  a BOOLEAN DEFAULT TRUE "
            "}"
            "END",
            'gser')

        datas = [
            ('A',             {'a': True}, b'a A ::= { a TRUE }'),
            ('A', {'a': False, 'b': True}, b'a A ::= { a FALSE, b TRUE }'),
            ('B',                      {}, b'b B ::= { }')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

        # Missing member.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('A', {})

        self.assertEqual(str(cm.exception),
                         "Member 'a' not found in {}.")

    def test_sequence_of(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE OF INTEGER "
            "END",
            'gser')

        datas = [
            ('A',                   [], b'a A ::= { }'),
            ('A',                  [1], b'a A ::= { 1 }'),
            ('A',               [1, 3], b'a A ::= { 1, 3 }')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_set(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SET { "
            "  a BOOLEAN, "
            "  b BOOLEAN OPTIONAL "
            "}"
            "B ::= SEQUENCE { "
            "  a BOOLEAN DEFAULT TRUE "
            "}"
            "END",
            'gser')

        datas = [
            ('A',             {'a': True}, b'a A ::= { a TRUE }'),
            ('A', {'a': False, 'b': True}, b'a A ::= { a FALSE, b TRUE }'),
            ('B',                      {}, b'b B ::= { }')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

        # Missing member.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('A', {})

        self.assertEqual(str(cm.exception),
                         "Member 'a' not found in {}.")

    def test_set_of(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SET OF INTEGER "
            "END",
            'gser')

        datas = [
            ('A',                   [], b'a A ::= { }'),
            ('A',                  [1], b'a A ::= { 1 }'),
            ('A',               [1, 3], b'a A ::= { 1, 3 }')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_choice(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= CHOICE { "
            "  a NULL "
            "} "
            "END",
            'gser')

        # Bad choice.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('A', ('b', None))

        self.assertEqual(str(cm.exception),
                         "Expected choice 'a', but got 'b'.")

    def test_utf8_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTF8String "
            "END",
            'gser')

        datas = [
            ('A',      'foo', b'a A ::= "foo"'),
            ('A',  u'\u0102', b'a A ::= "\xc4\x82"')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_numeric_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= NumericString "
            "END",
            'gser')

        decoded = '01 23'
        encoded = b'a A ::= "01 23"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_printable_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= PrintableString "
            "END",
            'gser')

        decoded = 'foo'
        encoded = b'a A ::= "foo"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_visible_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= VisibleString "
            "END",
            'gser')

        decoded = 'foo'
        encoded = b'a A ::= "foo"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_general_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GeneralString "
            "END",
            'gser')

        decoded = 'foo'
        encoded = b'a A ::= "foo"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_bmp_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BMPString "
            "END",
            'gser')

        datas = [
            ('A',      'foo', b'a A ::= "foo"'),
            ('A',  u'\u0102', b'a A ::= "\xc4\x82"')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_graphic_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GraphicString "
            "END",
            'gser')

        decoded = 'foo'
        encoded = b'a A ::= "foo"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_teletex_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= TeletexString "
            "END",
            'gser')

        decoded = 'foo'
        encoded = b'a A ::= "foo"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_universal_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UniversalString "
            "END",
            'gser')

        datas = [
            ('A',      u'bar', b'a A ::= "bar"')
        ]

        for type_name, decoded, encoded in datas:
            self.assertEqual(foo.encode(type_name, decoded), encoded)

    def test_utc_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTCTime "
            "END",
            'gser')

        decoded = ut2dt('010203040506Z')
        encoded = b'a A ::= "010203040506Z"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_generalized_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GeneralizedTime "
            "END",
            'gser')

        decoded = gt2dt('20001231235959.999Z')
        encoded = b'a A ::= "20001231235959.999Z"'

        self.assertEqual(foo.encode('A', decoded), encoded)

    def test_any(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= ANY "
            "END",
            'gser')

        datas = [
            ('A',                   b'', b"a A ::= ''H"),
            ('A',
             b'\x01\x23\x45\x67\x89\xab\xcd\xef',
             b"a A ::= '0123456789ABCDEF'H")
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_foo(self):
        foo = asn1tools.compile_files(['tests/files/foo.asn'], 'gser')

        datas = [
            ('Question',
             {'id': 1, 'question': 'Is 1+1=3?'},
             b'question Question ::= { id 1, question "Is 1+1=3?" }')
        ]

        for name, decoded, encoded in datas:
            self.assertEqual(foo.encode(name, decoded), encoded)

    def test_foo_indent(self):
        foo = asn1tools.compile_files(['tests/files/foo.asn'], 'gser')

        datas = [
            ('Question',
             {'id': 1, 'question': 'Is 1+1=3?'},
             b'question Question ::= {\n'
             b'  id 1,\n'
             b'  question "Is 1+1=3?"\n'
             b'}')
        ]

        for name, decoded, encoded in datas:
            actual = foo.encode(name, decoded, indent=2)
            self.assertEqual(actual, encoded)

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'gser')

        decoded = {
            'message': (
                'c1',
                (
                    'systemInformation',
                    {
                        'criticalExtensions': (
                            'systemInformation-r8',
                            {
                                'sib-TypeAndInfo': [
                                    (
                                        'sib2',
                                        {
                                            'ac-BarringInfo': {
                                                'ac-BarringForEmergency': True,
                                                'ac-BarringForMO-Data': {
                                                    'ac-BarringFactor': 'p95',
                                                    'ac-BarringTime': 's128',
                                                    'ac-BarringForSpecialAC': (b'\xf0', 5)
                                                }
                                            },
                                            'radioResourceConfigCommon': {
                                                'rach-ConfigCommon': {
                                                    'preambleInfo': {
                                                        'numberOfRA-Preambles': 'n24',
                                                        'preamblesGroupAConfig': {
                                                            'sizeOfRA-PreamblesGroupA': 'n28',
                                                            'messageSizeGroupA': 'b144',
                                                            'messagePowerOffsetGroupB': 'minusinfinity'
                                                        }
                                                    },
                                                    'powerRampingParameters': {
                                                        'powerRampingStep': 'dB0',
                                                        'preambleInitialReceivedTargetPower': 'dBm-102'
                                                    },
                                                    'ra-SupervisionInfo': {
                                                        'preambleTransMax': 'n8',
                                                        'ra-ResponseWindowSize': 'sf6',
                                                        'mac-ContentionResolutionTimer': 'sf48'
                                                    },
                                                    'maxHARQ-Msg3Tx': 8
                                                },
                                                'bcch-Config': {
                                                    'modificationPeriodCoeff': 'n2'
                                                },
                                                'pcch-Config': {
                                                    'defaultPagingCycle': 'rf256',
                                                    'nB': 'twoT'
                                                },
                                                'prach-Config': {
                                                    'rootSequenceIndex': 836,
                                                    'prach-ConfigInfo': {
                                                        'prach-ConfigIndex': 33,
                                                        'highSpeedFlag': False,
                                                        'zeroCorrelationZoneConfig': 10,
                                                        'prach-FreqOffset': 64
                                                    }
                                                },
                                                'pdsch-ConfigCommon': {
                                                    'referenceSignalPower': -60,
                                                    'p-b': 2
                                                },
                                                'pusch-ConfigCommon': {
                                                    'pusch-ConfigBasic': {
                                                        'n-SB': 1,
                                                        'hoppingMode': 'interSubFrame',
                                                        'pusch-HoppingOffset': 10,
                                                        'enable64QAM': False
                                                    },
                                                    'ul-ReferenceSignalsPUSCH': {
                                                        'groupHoppingEnabled': True,
                                                        'groupAssignmentPUSCH': 22,
                                                        'sequenceHoppingEnabled': False,
                                                        'cyclicShift': 5
                                                    }
                                                },
                                                'pucch-ConfigCommon': {
                                                    'deltaPUCCH-Shift': 'ds1',
                                                    'nRB-CQI': 98,
                                                    'nCS-AN': 4,
                                                    'n1PUCCH-AN': 2047
                                                },
                                                'soundingRS-UL-ConfigCommon': (
                                                    'setup',
                                                    {
                                                        'srs-BandwidthConfig': 'bw0',
                                                        'srs-SubframeConfig': 'sc4',
                                                        'ackNackSRS-SimultaneousTransmission': True
                                                    }),
                                                'uplinkPowerControlCommon': {
                                                    'p0-NominalPUSCH': -126,
                                                    'alpha': 'al0',
                                                    'p0-NominalPUCCH': -127,
                                                    'deltaFList-PUCCH': {
                                                        'deltaF-PUCCH-Format1': 'deltaF-2',
                                                        'deltaF-PUCCH-Format1b': 'deltaF1',
                                                        'deltaF-PUCCH-Format2': 'deltaF0',
                                                        'deltaF-PUCCH-Format2a': 'deltaF-2',
                                                        'deltaF-PUCCH-Format2b': 'deltaF0'
                                                    },
                                                    'deltaPreambleMsg3': -1
                                                },
                                                'ul-CyclicPrefixLength': 'len1'
                                            },
                                            'ue-TimersAndConstants': {
                                                't300': 'ms100',
                                                't301': 'ms200',
                                                't310': 'ms50',
                                                'n310': 'n2',
                                                't311': 'ms30000',
                                                'n311': 'n2'
                                            },
                                            'freqInfo': {
                                                'additionalSpectrumEmission': 3
                                            },
                                            'timeAlignmentTimerCommon': 'sf500'
                                        }
                                    ),
                                    (
                                        'sib3',
                                        {
                                            'cellReselectionInfoCommon': {
                                                'q-Hyst': 'dB0',
                                                'speedStateReselectionPars': {
                                                    'mobilityStateParameters': {
                                                        't-Evaluation': 's180',
                                                        't-HystNormal': 's180',
                                                        'n-CellChangeMedium': 1,
                                                        'n-CellChangeHigh': 16
                                                    },
                                                    'q-HystSF': {
                                                        'sf-Medium': 'dB-6',
                                                        'sf-High': 'dB-4'
                                                    }
                                                }
                                            },
                                            'cellReselectionServingFreqInfo': {
                                                'threshServingLow': 7,
                                                'cellReselectionPriority': 3
                                            },
                                            'intraFreqCellReselectionInfo': {
                                                'q-RxLevMin': -33,
                                                's-IntraSearch': 0,
                                                'presenceAntennaPort1': False,
                                                'neighCellConfig': (b'\x80', 2),
                                                't-ReselectionEUTRA': 4
                                            }
                                        }
                                    ),
                                    (
                                        'sib4',
                                        {
                                        }
                                    ),
                                    (
                                        'sib5',
                                        {
                                            'interFreqCarrierFreqList': [
                                                {
                                                    'dl-CarrierFreq': 1,
                                                    'q-RxLevMin': -45,
                                                    't-ReselectionEUTRA': 0,
                                                    'threshX-High': 31,
                                                    'threshX-Low': 29,
                                                    'allowedMeasBandwidth': 'mbw6',
                                                    'presenceAntennaPort1': True,
                                                    'neighCellConfig': (b'\x00', 2),
                                                    'q-OffsetFreq': 'dB0'
                                                }
                                            ]
                                        }
                                    ),
                                    (
                                        'sib6',
                                        {
                                            't-ReselectionUTRA': 3
                                        }
                                    ),
                                    ('sib7',
                                     {
                                         't-ReselectionGERAN': 3
                                     }
                                    ),
                                    ('sib8',
                                     {
                                         'parameters1XRTT': {
                                             'longCodeState1XRTT': (b'\x01#Eg\x89\x00', 42)
                                         }
                                     }
                                    ),
                                    (
                                        'sib9',
                                        {
                                            'hnb-Name': b'4'
                                        }
                                    ),
                                    ('sib10',
                                     {
                                         'messageIdentifier': (b'#4', 16),
                                         'serialNumber': (b'\x124', 16),
                                         'warningType': b'2\x12'
                                     }
                                    ),
                                    (
                                        'sib11',
                                        {
                                            'messageIdentifier': (b'g\x88', 16),
                                            'serialNumber': (b'T5', 16),
                                            'warningMessageSegmentType': 'notLastSegment',
                                            'warningMessageSegmentNumber': 19,
                                            'warningMessageSegment': b'\x12'
                                        }
                                    )
                                ]
                            }
                        )
                    }
                )
            )
        }

        encoded = (
            b"bcch-dl-sch-message BCCH-DL-SCH-Message ::= {\n"
            b"  message c1 : systemInformation : {\n"
            b"    criticalExtensions systemInformation-r8 : {\n"
            b"      sib-TypeAndInfo {\n"
            b"        sib2 : {\n"
            b"          ac-BarringInfo {\n"
            b"            ac-BarringForEmergency TRUE,\n"
            b"            ac-BarringForMO-Data {\n"
            b"              ac-BarringFactor p95,\n"
            b"              ac-BarringTime s128,\n"
            b"              ac-BarringForSpecialAC '11110'B\n"
            b"            }\n"
            b"          },\n"
            b"          radioResourceConfigCommon {\n"
            b"            rach-ConfigCommon {\n"
            b"              preambleInfo {\n"
            b"                numberOfRA-Preambles n24,\n"
            b"                preamblesGroupAConfig {\n"
            b"                  sizeOfRA-PreamblesGroupA n28,\n"
            b"                  messageSizeGroupA b144,\n"
            b"                  messagePowerOffsetGroupB minusinfinity\n"
            b"                }\n"
            b"              },\n"
            b"              powerRampingParameters {\n"
            b"                powerRampingStep dB0,\n"
            b"                preambleInitialReceivedTargetPower dBm-102\n"
            b"              },\n"
            b"              ra-SupervisionInfo {\n"
            b"                preambleTransMax n8,\n"
            b"                ra-ResponseWindowSize sf6,\n"
            b"                mac-ContentionResolutionTimer sf48\n"
            b"              },\n"
            b"              maxHARQ-Msg3Tx 8\n"
            b"            },\n"
            b"            bcch-Config {\n"
            b"              modificationPeriodCoeff n2\n"
            b"            },\n"
            b"            pcch-Config {\n"
            b"              defaultPagingCycle rf256,\n"
            b"              nB twoT\n"
            b"            },\n"
            b"            prach-Config {\n"
            b"              rootSequenceIndex 836,\n"
            b"              prach-ConfigInfo {\n"
            b"                prach-ConfigIndex 33,\n"
            b"                highSpeedFlag FALSE,\n"
            b"                zeroCorrelationZoneConfig 10,\n"
            b"                prach-FreqOffset 64\n"
            b"              }\n"
            b"            },\n"
            b"            pdsch-ConfigCommon {\n"
            b"              referenceSignalPower -60,\n"
            b"              p-b 2\n"
            b"            },\n"
            b"            pusch-ConfigCommon {\n"
            b"              pusch-ConfigBasic {\n"
            b"                n-SB 1,\n"
            b"                hoppingMode interSubFrame,\n"
            b"                pusch-HoppingOffset 10,\n"
            b"                enable64QAM FALSE\n"
            b"              },\n"
            b"              ul-ReferenceSignalsPUSCH {\n"
            b"                groupHoppingEnabled TRUE,\n"
            b"                groupAssignmentPUSCH 22,\n"
            b"                sequenceHoppingEnabled FALSE,\n"
            b"                cyclicShift 5\n"
            b"              }\n"
            b"            },\n"
            b"            pucch-ConfigCommon {\n"
            b"              deltaPUCCH-Shift ds1,\n"
            b"              nRB-CQI 98,\n"
            b"              nCS-AN 4,\n"
            b"              n1PUCCH-AN 2047\n"
            b"            },\n"
            b"            soundingRS-UL-ConfigCommon setup : {\n"
            b"              srs-BandwidthConfig bw0,\n"
            b"              srs-SubframeConfig sc4,\n"
            b"              ackNackSRS-SimultaneousTransmission TRUE\n"
            b"            },\n"
            b"            uplinkPowerControlCommon {\n"
            b"              p0-NominalPUSCH -126,\n"
            b"              alpha al0,\n"
            b"              p0-NominalPUCCH -127,\n"
            b"              deltaFList-PUCCH {\n"
            b"                deltaF-PUCCH-Format1 deltaF-2,\n"
            b"                deltaF-PUCCH-Format1b deltaF1,\n"
            b"                deltaF-PUCCH-Format2 deltaF0,\n"
            b"                deltaF-PUCCH-Format2a deltaF-2,\n"
            b"                deltaF-PUCCH-Format2b deltaF0\n"
            b"              },\n"
            b"              deltaPreambleMsg3 -1\n"
            b"            },\n"
            b"            ul-CyclicPrefixLength len1\n"
            b"          },\n"
            b"          ue-TimersAndConstants {\n"
            b"            t300 ms100,\n"
            b"            t301 ms200,\n"
            b"            t310 ms50,\n"
            b"            n310 n2,\n"
            b"            t311 ms30000,\n"
            b"            n311 n2\n"
            b"          },\n"
            b"          freqInfo {\n"
            b"            additionalSpectrumEmission 3\n"
            b"          },\n"
            b"          timeAlignmentTimerCommon sf500\n"
            b"        },\n"
            b"        sib3 : {\n"
            b"          cellReselectionInfoCommon {\n"
            b"            q-Hyst dB0,\n"
            b"            speedStateReselectionPars {\n"
            b"              mobilityStateParameters {\n"
            b"                t-Evaluation s180,\n"
            b"                t-HystNormal s180,\n"
            b"                n-CellChangeMedium 1,\n"
            b"                n-CellChangeHigh 16\n"
            b"              },\n"
            b"              q-HystSF {\n"
            b"                sf-Medium dB-6,\n"
            b"                sf-High dB-4\n"
            b"              }\n"
            b"            }\n"
            b"          },\n"
            b"          cellReselectionServingFreqInfo {\n"
            b"            threshServingLow 7,\n"
            b"            cellReselectionPriority 3\n"
            b"          },\n"
            b"          intraFreqCellReselectionInfo {\n"
            b"            q-RxLevMin -33,\n"
            b"            s-IntraSearch 0,\n"
            b"            presenceAntennaPort1 FALSE,\n"
            b"            neighCellConfig '10'B,\n"
            b"            t-ReselectionEUTRA 4\n"
            b"          }\n"
            b"        },\n"
            b"        sib4 : {\n"
            b"        },\n"
            b"        sib5 : {\n"
            b"          interFreqCarrierFreqList {\n"
            b"            {\n"
            b"              dl-CarrierFreq 1,\n"
            b"              q-RxLevMin -45,\n"
            b"              t-ReselectionEUTRA 0,\n"
            b"              threshX-High 31,\n"
            b"              threshX-Low 29,\n"
            b"              allowedMeasBandwidth mbw6,\n"
            b"              presenceAntennaPort1 TRUE,\n"
            b"              neighCellConfig '00'B,\n"
            b"              q-OffsetFreq dB0\n"
            b"            }\n"
            b"          }\n"
            b"        },\n"
            b"        sib6 : {\n"
            b"          t-ReselectionUTRA 3\n"
            b"        },\n"
            b"        sib7 : {\n"
            b"          t-ReselectionGERAN 3\n"
            b"        },\n"
            b"        sib8 : {\n"
            b"          parameters1XRTT {\n"
            b"            longCodeState1XRTT '000000010010001101000101011001111000100100'B\n"
            b"          }\n"
            b"        },\n"
            b"        sib9 : {\n"
            b"          hnb-Name '34'H\n"
            b"        },\n"
            b"        sib10 : {\n"
            b"          messageIdentifier '0010001100110100'B,\n"
            b"          serialNumber '0001001000110100'B,\n"
            b"          warningType '3212'H\n"
            b"        },\n"
            b"        sib11 : {\n"
            b"          messageIdentifier '0110011110001000'B,\n"
            b"          serialNumber '0101010000110101'B,\n"
            b"          warningMessageSegmentType notLastSegment,\n"
            b"          warningMessageSegmentNumber 19,\n"
            b"          warningMessageSegment '12'H\n"
            b"        }\n"
            b"      }\n"
            b"    }\n"
            b"  }\n"
            b"}"
        )

        self.assertEqual(rrc.encode('BCCH-DL-SCH-Message', decoded, indent=2),
                         encoded)

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'gser')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Real']), 'Real(Real)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Graphicstring']),
                         'GraphicString(Graphicstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')
        self.assertEqual(repr(all_types.types['GeneralizedTime1']),
                         'GeneralizedTime(GeneralizedTime1)')
        self.assertEqual(repr(all_types.types['Choice']),
                         'Choice(Choice, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Any']), 'Any(Any)')
        self.assertEqual(repr(all_types.types['Sequence12']),
                         'Sequence(Sequence12, [SequenceOf(a, Recursive(Sequence12))])')

    def test_rfc4511(self):
        rfc4511 = asn1tools.compile_dict(deepcopy(RFC4511), 'gser')

        # A search request message.
        decoded = {
            'messageID': 2,
            'protocolOp': (
                'searchRequest',
                {
                    'baseObject': b'',
                    'scope': 'wholeSubtree',
                    'derefAliases': 'neverDerefAliases',
                    'sizeLimit': 0,
                    'timeLimit': 0,
                    'typesOnly': False,
                    'filter': (
                        'and',
                        [
                            (
                                'substrings',
                                {
                                    'type': b'\x63\x6e',
                                    'substrings': [
                                        ('any', b'\x66\x72\x65\x64')
                                    ]
                                }
                            ),
                            (
                                'equalityMatch',
                                {
                                    'attributeDesc': b'\x64\x6e',
                                    'assertionValue': b'\x6a\x6f\x65'
                                }
                            )
                        ]
                    ),
                    'attributes': [
                    ]
                }
            )
        }

        encoded = (
            b"ldapmessage LDAPMessage ::= {\n"
            b"  messageID 2,\n"
            b"  protocolOp searchRequest : {\n"
            b"    baseObject ''H,\n"
            b"    scope wholeSubtree,\n"
            b"    derefAliases neverDerefAliases,\n"
            b"    sizeLimit 0,\n"
            b"    timeLimit 0,\n"
            b"    typesOnly FALSE,\n"
            b"    filter and : {\n"
            b"      substrings : {\n"
            b"        type '636E'H,\n"
            b"        substrings {\n"
            b"          any : '66726564'H\n"
            b"        }\n"
            b"      },\n"
            b"      equalityMatch : {\n"
            b"        attributeDesc '646E'H,\n"
            b"        assertionValue '6A6F65'H\n"
            b"      }\n"
            b"    },\n"
            b"    attributes {\n"
            b"    }\n"
            b"  }\n"
            b"}"
        )

        self.assertEqual(rfc4511.encode('LDAPMessage', decoded, indent=2),
                         encoded)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'RFC1157-SNMP': {'extensibility-implied': False,
                  'imports': {'RFC1155-SMI': ['ObjectName',
                                              'ObjectSyntax',
                                              'NetworkAddress',
                                              'IpAddress',
                                              'TimeTicks']},
                  'object-classes': {},
                  'object-sets': {},
                  'types': {'GetNextRequest-PDU': {'tag': {'kind': 'IMPLICIT',
                                                           'number': 1},
                                                   'type': 'PDU'},
                            'GetRequest-PDU': {'tag': {'kind': 'IMPLICIT',
                                                       'number': 0},
                                               'type': 'PDU'},
                            'GetResponse-PDU': {'tag': {'kind': 'IMPLICIT',
                                                        'number': 2},
                                                'type': 'PDU'},
                            'Message': {'members': [{'name': 'version',
                                                     'type': 'INTEGER'},
                                                    {'name': 'community',
                                                     'type': 'OCTET STRING'},
                                                    {'name': 'data',
                                                     'type': 'PDUs'}],
                                        'type': 'SEQUENCE'},
                            'PDU': {'members': [{'name': 'request-id',
                                                 'type': 'INTEGER'},
                                                {'name': 'error-status',
                                                 'type': 'INTEGER'},
                                                {'name': 'error-index',
                                                 'type': 'INTEGER'},
                                                {'name': 'variable-bindings',
                                                 'type': 'VarBindList'}],
                                    'type': 'SEQUENCE'},
                            'PDUs': {'members': [{'name': 'get-request',
                                                  'type': 'GetRequest-PDU'},
                                                 {'name': 'get-next-request',
                                                  'type': 'GetNextRequest-PDU'},
                                                 {'name': 'get-response',
                                                  'type': 'GetResponse-PDU'},
                                                 {'name': 'set-request',
                                                  'type': 'SetRequest-PDU'},
                                                 {'name': 'trap',
                                                  'type': 'Trap-PDU'}],
                                     'type': 'CHOICE'},
                            'SetRequest-PDU': {'tag': {'kind': 'IMPLICIT',
                                                       'number': 3},
                                               'type': 'PDU'},
                            'Trap-PDU': {'members': [{'name': 'enterprise',
                                                      'type': 'OBJECT '
                                                              'IDENTIFIER'},
                                                     {'name': 'agent-addr',
                                                      'type': 'NetworkAddress'},
                                                     {'name': 'generic-trap',
                                                      'type': 'INTEGER'},
                                                     {'name': 'specific-trap',
                                                      'type': 'INTEGER'},
                                                     {'name': 'time-stamp',
                                                      'type': 'TimeTicks'},
                                                     {'name': 'variable-bindings',
                                                      'type': 'VarBindList'}],
                                         'tag': {'kind': 'IMPLICIT',
                                                 'number': 4},
                                         'type': 'SEQUENCE'},
                            'VarBind': {'members': [{'name': 'name',
                                                     'type': 'ObjectName'},
                                                    {'name': 'value',
                                                     'type': 'ObjectSyntax'}],
                                        'type': 'SEQUENCE'},
                            'VarBindList': {'element': {'type': 'VarBind'},
                                            'type': 'SEQUENCE OF'}},
                  'values': {}}}


EXPECTED = {'AllTypes': {'extensibility-implied': False,
              'imports': {},
              'object-classes': {},
              'object-sets': {},
              'tags': 'AUTOMATIC',
              'types': {'Sequence2': {'members': [{'name': 'a',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 0},
                                                   'type': 'INTEGER'},
                                                  {'name': 'b',
                                                   'optional': True,
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 1},
                                                   'type': 'BOOLEAN'},
                                                  {'name': 'c',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 2},
                                                   'type': 'INTEGER'}],
                                      'type': 'SEQUENCE'},
                        'Sequence3': {'members': [{'name': 'a',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 0},
                                                   'type': 'INTEGER'},
                                                  {'name': 'b',
                                                   'optional': True,
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 1},
                                                   'type': 'BOOLEAN'},
                                                  {'name': 'c',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 2},
                                                   'type': 'INTEGER'},
                                                  {'name': 'd',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 3},
                                                   'type': 'BOOLEAN'}],
                                      'type': 'SEQUENCE'},
                        'Sequence9': {'members': [{'name': 'a',
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 0},
                                                   'type': 'INTEGER'},
                                                  {'name': 'b',
                                                   'optional': True,
                                                   'tag': {'kind': 'IMPLICIT',
                                                           'number': 1},
                                                   'type': 'BOOLEAN'}],
                                      'type': 'SEQUENCE'}},
              'values': {}}}



"""The top level of the ASN.1 tools package contains commonly used
functions and classes, and the command line interface.

"""

import sys
import os
import argparse
import binascii
import logging
from pprint import pformat
import pickle

from prompt_toolkit.completion import WordCompleter
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from .compiler import compile_dict
from .compiler import compile_string
from .compiler import compile_files
from .compiler import pre_process_dict
from .parser import parse_string
from .parser import parse_files
from .parser import ParseError
from .errors import Error
from .errors import EncodeError
from .errors import DecodeError
from .errors import CompileError
from .errors import ConstraintsError


__author__ = 'Erik Moqvist'
__version__ = '0.116.0'


class ArgumentParserError(Error):
    pass


class ArgumentParser(argparse.ArgumentParser):

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stdout)

        raise ArgumentParserError(message)

    def error(self, message):
        self.print_usage(sys.stdout)
        message = '{}: error: {}'.format(self.prog, message)
        print(message)

        raise ArgumentParserError(message)


def _import_module(pyfilepath):
    module_name = os.path.splitext(os.path.basename(pyfilepath))[0]

    if sys.version_info > (3, 4):
        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, pyfilepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        import imp
        module = imp.load_source(module_name, pyfilepath)

    return module


def _convert_hexstring(input_spec,
                       output_spec,
                       output_codec,
                       type_name,
                       hexstring):
    try:
        encoded = binascii.unhexlify(hexstring)
    except Exception as e:
        raise TypeError("'{}': {}".format(hexstring, str(e)))

    decoded = input_spec.decode(type_name, encoded)

    if output_codec in ['gser', 'xer', 'jer']:
        decoded = output_spec.encode(type_name, decoded, indent=4).strip()
    else:
        decoded = binascii.hexlify(output_spec.encode(type_name, decoded))

    print(decoded.decode('latin-1'))


def _compile_files(specs, input_codec, output_codec):
    py_count = [spec.endswith('.py') for spec in specs].count(True)
    pkl_count = [spec.endswith('.pkl') for spec in specs].count(True)

    if py_count > 0:
        if py_count != 1:
            raise Exception('Expected one .py-file, but got {}.'.format(py_count))

        module = _import_module(specs[0])
        parsed = module.SPECIFICATION
        input_spec = compile_dict(parsed, input_codec)
        output_spec = compile_dict(parsed, output_codec)
    elif pkl_count > 0:
        if pkl_count != 2:
            raise Exception('Expected two .pkl-files, but got {}.'.format(pkl_count))

        with open(specs[0], 'rb') as fin:
            input_spec = pickle.load(fin)

        with open(specs[1], 'rb') as fin:
            output_spec = pickle.load(fin)
    else:
        parsed = parse_files(specs)
        input_spec = compile_dict(parsed, input_codec)
        output_spec = compile_dict(parsed, output_codec)

    return input_spec, output_spec


def _do_convert(args):
    input_spec, output_spec = _compile_files(args.specification,
                                             args.input_codec,
                                             args.output_codec)

    if args.hexstring == '-':
        for hexstring in sys.stdin:
            hexstring = hexstring.strip('\r\n')

            if hexstring:
                try:
                    _convert_hexstring(input_spec,
                                       output_spec,
                                       args.output_codec,
                                       args.type,
                                       hexstring)
                except TypeError:
                    print(hexstring)
                except DecodeError as e:
                    print(hexstring)
                    print(str(e))
            else:
                print(hexstring)
    else:
        _convert_hexstring(input_spec,
                           output_spec,
                           args.output_codec,
                           args.type,
                           args.hexstring)


def _handle_command_compile(line):
    parser = ArgumentParser(prog='compile')
    parser.add_argument('-i', '--input-codec',
                        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer'),
                        default='ber',
                        help='Input codec (default: ber).')
    parser.add_argument('-o', '--output-codec',
                        choices=(
                            'ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer', 'gser'
                        ),
                        default='gser',
                        help='Output codec (default: gser).')
    parser.add_argument('specification',
                        nargs='+',
                        help='ASN.1 specification as one or more .asn files.')

    try:
        args = parser.parse_args(args=line.split()[1:])
    except ArgumentParserError:
        return None, None, None

    try:
        input_spec, output_spec = _compile_files(args.specification,
                                                 args.input_codec,
                                                 args.output_codec)
        return input_spec, output_spec, args.output_codec
    except Exception as e:
        print('error: {}'.format(str(e)))

        return None, None, None


def _handle_command_convert(line, input_spec, output_spec, output_codec):
    if not input_spec:
        print("No compiled specification found. Please use the "
              "'compile' command to compile one.")
        return

    parser = ArgumentParser(prog='convert')
    parser.add_argument('type', help='Type to convert.')
    parser.add_argument(
        'hexstring',
        help='Hexstring to convert.')

    try:
        args = parser.parse_args(args=line.split()[1:])
    except ArgumentParserError:
        return

    try:
        _convert_hexstring(input_spec,
                           output_spec,
                           output_codec,
                           args.type,
                           args.hexstring)
    except Exception as e:
        print('error: {}'.format(str(e)))


def _handle_command_help():
    print('Commands:')
    print('  compile')
    print('  convert')
    print('  exit')
    print('  help')


def _do_shell(_args):
    commands = ['compile', 'convert', 'help', 'exit']
    completer = WordCompleter(commands, WORD=True)
    user_home = os.path.expanduser('~')
    history = FileHistory(os.path.join(user_home, '.asn1tools-history.txt'))
    session = PromptSession(completer=completer,
                            complete_while_typing=True,
                            auto_suggest=AutoSuggestFromHistory(),
                            enable_history_search=True,
                            history=history)
    input_spec = None
    output_spec = None
    output_codec = None

    print("\nWelcome to the asn1tools shell!\n")

    while True:
        try:
            line = session.prompt(u'$ ')
        except EOFError:
            return

        line = line.strip()

        if line:
            if line.startswith('compile'):
                input_spec, output_spec, output_codec = _handle_command_compile(line)
            elif line.startswith('convert'):
                _handle_command_convert(line,
                                        input_spec,
                                        output_spec,
                                        output_codec)
            elif line == 'help':
                _handle_command_help()
            elif line == 'exit':
                return
            else:
                print('{}: command not found'.format(line))


def _do_parse(args):
    parsed = parse_files(args.specification)

    with open(args.outfile, 'w') as fout:
        fout.write('SPECIFICATION = {}'.format(pformat(parsed)))


def _do_pickle(args):
    compiled = compile_files(args.specification, args.codec)

    with open(args.outfile, 'wb') as fout:
        pickle.dump(compiled, fout)


def _main():
    parser = argparse.ArgumentParser(
        description='Various ASN.1 utilities.')

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose',
                        choices=(0, 1, 2),
                        type=int,
                        default=1,
                        help=('Control the verbosity; disable(0), warning(1) '
                              'and debug(2). (default: 1).'))
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # The 'convert' subparser.
    subparser = subparsers.add_parser(
        'convert',
        description='Convert given hextring and print it to standard output.')
    subparser.add_argument(
        '-i', '--input-codec',
        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer'),
        default='ber',
        help='Input format (default: ber).')
    subparser.add_argument(
        '-o', '--output-codec',
        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer', 'gser'),
        default='gser',
        help='Output format (default: gser).')
    subparser.add_argument(
        'specification',
        nargs='+',
        help=('ASN.1 specification as one or more .asn files, one .py file or '
              'one input and one output codec .pkl files. The .py-file may be '
              'created with the parse subcommand. The .pkl-files may be '
              'created with the pickle subcommand.'))
    subparser.add_argument('type', help='Type to convert.')
    subparser.add_argument(
        'hexstring',
        help='Hexstring to convert, or - to read hexstrings from standard input.')
    subparser.set_defaults(func=_do_convert)

    # The 'shell' subparser.
    subparser = subparsers.add_parser('shell',
                                      description='An interactive shell.')
    subparser.set_defaults(func=_do_shell)

    # The 'parse' subparser.
    subparser = subparsers.add_parser('parse',
                                      description='Convert to a Python dictionary.')
    subparser.add_argument('specification',
                           nargs='+',
                           help='ASN.1 specification as one or more .asn files.')
    subparser.add_argument('outfile',
                           help='Output file name.')
    subparser.set_defaults(func=_do_parse)

    # The 'pickle' subparser.
    subparser = subparsers.add_parser('pickle',
                                      description='Pickle.')
    subparser.add_argument(
        '-c', '--codec',
        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer', 'gser'),
        default='ber',
        help='Codec (default: ber).')
    subparser.add_argument('specification',
                           nargs='+',
                           help='ASN.1 specification as one or more .asn files.')
    subparser.add_argument('outfile',
                           help='Output file name.')
    subparser.set_defaults(func=_do_pickle)

    args = parser.parse_args()

    levels = [logging.CRITICAL, logging.WARNING, logging.DEBUG]
    level = levels[args.verbose]

    logging.basicConfig(format='%(levelname)s: %(message)s',
                        level=level)

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit('error: {}'.format(str(e)))



EXPECTED = {'AttributeCertificateVersion1': {'extensibility-implied': False,
                                  'imports': {'PKIX1Explicit88': ['AlgorithmIdentifier',
                                                                  'Attribute',
                                                                  'CertificateSerialNumber',
                                                                  'Extensions',
                                                                  'UniqueIdentifier'],
                                              'PKIX1Implicit88': ['GeneralNames'],
                                              'PKIXAttributeCertificate': ['AttCertValidityPeriod',
                                                                           'IssuerSerial']},
                                  'object-classes': {},
                                  'object-sets': {},
                                  'tags': 'EXPLICIT',
                                  'types': {'AttCertVersionV1': {'type': 'INTEGER'},
                                            'AttributeCertificateInfoV1': {'members': [{'default': 'v1',
                                                                                        'name': 'version',
                                                                                        'type': 'AttCertVersionV1'},
                                                                                       {'members': [{'name': 'baseCertificateID',
                                                                                                     'tag': {'number': 0},
                                                                                                     'type': 'IssuerSerial'},
                                                                                                    {'name': 'subjectName',
                                                                                                     'tag': {'number': 1},
                                                                                                     'type': 'GeneralNames'}],
                                                                                        'name': 'subject',
                                                                                        'type': 'CHOICE'},
                                                                                       {'name': 'issuer',
                                                                                        'type': 'GeneralNames'},
                                                                                       {'name': 'signature',
                                                                                        'type': 'AlgorithmIdentifier'},
                                                                                       {'name': 'serialNumber',
                                                                                        'type': 'CertificateSerialNumber'},
                                                                                       {'name': 'attCertValidityPeriod',
                                                                                        'type': 'AttCertValidityPeriod'},
                                                                                       {'element': {'type': 'Attribute'},
                                                                                        'name': 'attributes',
                                                                                        'type': 'SEQUENCE '
                                                                                                'OF'},
                                                                                       {'name': 'issuerUniqueID',
                                                                                        'optional': True,
                                                                                        'type': 'UniqueIdentifier'},
                                                                                       {'name': 'extensions',
                                                                                        'optional': True,
                                                                                        'type': 'Extensions'}],
                                                                           'type': 'SEQUENCE'},
                                            'AttributeCertificateV1': {'members': [{'name': 'acInfo',
                                                                                    'type': 'AttributeCertificateInfoV1'},
                                                                                   {'name': 'signatureAlgorithm',
                                                                                    'type': 'AlgorithmIdentifier'},
                                                                                   {'name': 'signature',
                                                                                    'type': 'BIT '
                                                                                            'STRING'}],
                                                                       'type': 'SEQUENCE'}},
                                  'values': {}},
 'CryptographicMessageSyntax2004': {'extensibility-implied': False,
                                    'imports': {'AttributeCertificateVersion1': ['AttributeCertificateV1'],
                                                'PKIX1Explicit88': ['AlgorithmIdentifier',
                                                                    'Certificate',
                                                                    'CertificateList',
                                                                    'CertificateSerialNumber',
                                                                    'Name'],
                                                'PKIXAttributeCertificate': ['AttributeCertificate']},
                                    'object-classes': {},
                                    'object-sets': {},
                                    'tags': 'IMPLICIT',
                                    'types': {'Attribute': {'members': [{'name': 'attrType',
                                                                         'type': 'OBJECT '
                                                                                 'IDENTIFIER'},
                                                                        {'element': {'type': 'AttributeValue'},
                                                                         'name': 'attrValues',
                                                                         'type': 'SET '
                                                                                 'OF'}],
                                                            'type': 'SEQUENCE'},
                                              'AttributeCertificateV2': {'type': 'AttributeCertificate'},
                                              'AttributeValue': {'type': 'ANY'},
                                              'AuthAttributes': {'element': {'type': 'Attribute'},
                                                                 'size': [(1,
                                                                           'MAX')],
                                                                 'type': 'SET '
                                                                         'OF'},
                                              'AuthenticatedData': {'members': [{'name': 'version',
                                                                                 'type': 'CMSVersion'},
                                                                                {'name': 'originatorInfo',
                                                                                 'optional': True,
                                                                                 'tag': {'kind': 'IMPLICIT',
                                                                                         'number': 0},
                                                                                 'type': 'OriginatorInfo'},
                                                                                {'name': 'recipientInfos',
                                                                                 'type': 'RecipientInfos'},
                                                                                {'name': 'macAlgorithm',
                                                                                 'type': 'MessageAuthenticationCodeAlgorithm'},
                                                                                {'name': 'digestAlgorithm',
                                                                                 'optional': True,
                                                                                 'tag': {'number': 1},
                                                                                 'type': 'DigestAlgorithmIdentifier'},
                                                                                {'name': 'encapContentInfo',
                                                                                 'type': 'EncapsulatedContentInfo'},
                                                                                {'name': 'authAttrs',
                                                                                 'optional': True,
                                                                                 'tag': {'kind': 'IMPLICIT',
                                                                                         'number': 2},
                                                                                 'type': 'AuthAttributes'},
                                                                                {'name': 'mac',
                                                                                 'type': 'MessageAuthenticationCode'},
                                                                                {'name': 'unauthAttrs',
                                                                                 'optional': True,
                                                                                 'tag': {'kind': 'IMPLICIT',
                                                                                         'number': 3},
                                                                                 'type': 'UnauthAttributes'}],
                                                                    'type': 'SEQUENCE'},
                                              'CMSVersion': {'type': 'INTEGER'},
                                              'CertificateChoices': {'members': [{'name': 'certificate',
                                                                                  'type': 'Certificate'},
                                                                                 {'name': 'extendedCertificate',
                                                                                  'tag': {'kind': 'IMPLICIT',
                                                                                          'number': 0},
                                                                                  'type': 'ExtendedCertificate'},
                                                                                 {'name': 'v1AttrCert',
                                                                                  'tag': {'kind': 'IMPLICIT',
                                                                                          'number': 1},
                                                                                  'type': 'AttributeCertificateV1'},
                                                                                 {'name': 'v2AttrCert',
                                                                                  'tag': {'kind': 'IMPLICIT',
                                                                                          'number': 2},
                                                                                  'type': 'AttributeCertificateV2'},
                                                                                 {'name': 'other',
                                                                                  'tag': {'kind': 'IMPLICIT',
                                                                                          'number': 3},
                                                                                  'type': 'OtherCertificateFormat'}],
                                                                     'type': 'CHOICE'},
                                              'CertificateSet': {'element': {'type': 'CertificateChoices'},
                                                                 'type': 'SET '
                                                                         'OF'},
                                              'ContentEncryptionAlgorithmIdentifier': {'type': 'AlgorithmIdentifier'},
                                              'ContentInfo': {'members': [{'name': 'contentType',
                                                                           'type': 'ContentType'},
                                                                          {'choices': {},
                                                                           'name': 'content',
                                                                           'tag': {'kind': 'EXPLICIT',
                                                                                   'number': 0},
                                                                           'type': 'ANY '
                                                                                   'DEFINED '
                                                                                   'BY',
                                                                           'value': 'contentType'}],
                                                              'type': 'SEQUENCE'},
                                              'ContentType': {'type': 'OBJECT '
                                                                      'IDENTIFIER'},
                                              'Countersignature': {'type': 'SignerInfo'},
                                              'Digest': {'type': 'OCTET '
                                                                 'STRING'},
                                              'DigestAlgorithmIdentifier': {'type': 'AlgorithmIdentifier'},
                                              'DigestAlgorithmIdentifiers': {'element': {'type': 'DigestAlgorithmIdentifier'},
                                                                             'type': 'SET '
                                                                                     'OF'},
                                              'DigestedData': {'members': [{'name': 'version',
                                                                            'type': 'CMSVersion'},
                                                                           {'name': 'digestAlgorithm',
                                                                            'type': 'DigestAlgorithmIdentifier'},
                                                                           {'name': 'encapContentInfo',
                                                                            'type': 'EncapsulatedContentInfo'},
                                                                           {'name': 'digest',
                                                                            'type': 'Digest'}],
                                                               'type': 'SEQUENCE'},
                                              'EncapsulatedContentInfo': {'members': [{'name': 'eContentType',
                                                                                       'type': 'ContentType'},
                                                                                      {'name': 'eContent',
                                                                                       'optional': True,
                                                                                       'tag': {'kind': 'EXPLICIT',
                                                                                               'number': 0},
                                                                                       'type': 'OCTET '
                                                                                               'STRING'}],
                                                                          'type': 'SEQUENCE'},
                                              'EncryptedContent': {'type': 'OCTET '
                                                                           'STRING'},
                                              'EncryptedContentInfo': {'members': [{'name': 'contentType',
                                                                                    'type': 'ContentType'},
                                                                                   {'name': 'contentEncryptionAlgorithm',
                                                                                    'type': 'ContentEncryptionAlgorithmIdentifier'},
                                                                                   {'name': 'encryptedContent',
                                                                                    'optional': True,
                                                                                    'tag': {'kind': 'IMPLICIT',
                                                                                            'number': 0},
                                                                                    'type': 'EncryptedContent'}],
                                                                       'type': 'SEQUENCE'},
                                              'EncryptedData': {'members': [{'name': 'version',
                                                                             'type': 'CMSVersion'},
                                                                            {'name': 'encryptedContentInfo',
                                                                             'type': 'EncryptedContentInfo'},
                                                                            {'name': 'unprotectedAttrs',
                                                                             'optional': True,
                                                                             'tag': {'kind': 'IMPLICIT',
                                                                                     'number': 1},
                                                                             'type': 'UnprotectedAttributes'}],
                                                                'type': 'SEQUENCE'},
                                              'EncryptedKey': {'type': 'OCTET '
                                                                       'STRING'},
                                              'EnvelopedData': {'members': [{'name': 'version',
                                                                             'type': 'CMSVersion'},
                                                                            {'name': 'originatorInfo',
                                                                             'optional': True,
                                                                             'tag': {'kind': 'IMPLICIT',
                                                                                     'number': 0},
                                                                             'type': 'OriginatorInfo'},
                                                                            {'name': 'recipientInfos',
                                                                             'type': 'RecipientInfos'},
                                                                            {'name': 'encryptedContentInfo',
                                                                             'type': 'EncryptedContentInfo'},
                                                                            {'name': 'unprotectedAttrs',
                                                                             'optional': True,
                                                                             'tag': {'kind': 'IMPLICIT',
                                                                                     'number': 1},
                                                                             'type': 'UnprotectedAttributes'}],
                                                                'type': 'SEQUENCE'},
                                              'ExtendedCertificate': {'members': [{'name': 'extendedCertificateInfo',
                                                                                   'type': 'ExtendedCertificateInfo'},
                                                                                  {'name': 'signatureAlgorithm',
                                                                                   'type': 'SignatureAlgorithmIdentifier'},
                                                                                  {'name': 'signature',
                                                                                   'type': 'Signature'}],
                                                                      'type': 'SEQUENCE'},
                                              'ExtendedCertificateInfo': {'members': [{'name': 'version',
                                                                                       'type': 'CMSVersion'},
                                                                                      {'name': 'certificate',
                                                                                       'type': 'Certificate'},
                                                                                      {'name': 'attributes',
                                                                                       'type': 'UnauthAttributes'}],
                                                                          'type': 'SEQUENCE'},
                                              'ExtendedCertificateOrCertificate': {'members': [{'name': 'certificate',
                                                                                                'type': 'Certificate'},
                                                                                               {'name': 'extendedCertificate',
                                                                                                'tag': {'kind': 'IMPLICIT',
                                                                                                        'number': 0},
                                                                                                'type': 'ExtendedCertificate'}],
                                                                                   'type': 'CHOICE'},
                                              'IssuerAndSerialNumber': {'members': [{'name': 'issuer',
                                                                                     'type': 'Name'},
                                                                                    {'name': 'serialNumber',
                                                                                     'type': 'CertificateSerialNumber'}],
                                                                        'type': 'SEQUENCE'},
                                              'KEKIdentifier': {'members': [{'name': 'keyIdentifier',
                                                                             'type': 'OCTET '
                                                                                     'STRING'},
                                                                            {'name': 'date',
                                                                             'optional': True,
                                                                             'type': 'GeneralizedTime'},
                                                                            {'name': 'other',
                                                                             'optional': True,
                                                                             'type': 'OtherKeyAttribute'}],
                                                                'type': 'SEQUENCE'},
                                              'KEKRecipientInfo': {'members': [{'name': 'version',
                                                                                'type': 'CMSVersion'},
                                                                               {'name': 'kekid',
                                                                                'type': 'KEKIdentifier'},
                                                                               {'name': 'keyEncryptionAlgorithm',
                                                                                'type': 'KeyEncryptionAlgorithmIdentifier'},
                                                                               {'name': 'encryptedKey',
                                                                                'type': 'EncryptedKey'}],
                                                                   'type': 'SEQUENCE'},
                                              'KeyAgreeRecipientIdentifier': {'members': [{'name': 'issuerAndSerialNumber',
                                                                                           'type': 'IssuerAndSerialNumber'},
                                                                                          {'name': 'rKeyId',
                                                                                           'tag': {'kind': 'IMPLICIT',
                                                                                                   'number': 0},
                                                                                           'type': 'RecipientKeyIdentifier'}],
                                                                              'type': 'CHOICE'},
                                              'KeyAgreeRecipientInfo': {'members': [{'name': 'version',
                                                                                     'type': 'CMSVersion'},
                                                                                    {'name': 'originator',
                                                                                     'tag': {'kind': 'EXPLICIT',
                                                                                             'number': 0},
                                                                                     'type': 'OriginatorIdentifierOrKey'},
                                                                                    {'name': 'ukm',
                                                                                     'optional': True,
                                                                                     'tag': {'kind': 'EXPLICIT',
                                                                                             'number': 1},
                                                                                     'type': 'UserKeyingMaterial'},
                                                                                    {'name': 'keyEncryptionAlgorithm',
                                                                                     'type': 'KeyEncryptionAlgorithmIdentifier'},
                                                                                    {'name': 'recipientEncryptedKeys',
                                                                                     'type': 'RecipientEncryptedKeys'}],
                                                                        'type': 'SEQUENCE'},
                                              'KeyDerivationAlgorithmIdentifier': {'type': 'AlgorithmIdentifier'},
                                              'KeyEncryptionAlgorithmIdentifier': {'type': 'AlgorithmIdentifier'},
                                              'KeyTransRecipientInfo': {'members': [{'name': 'version',
                                                                                     'type': 'CMSVersion'},
                                                                                    {'name': 'rid',
                                                                                     'type': 'RecipientIdentifier'},
                                                                                    {'name': 'keyEncryptionAlgorithm',
                                                                                     'type': 'KeyEncryptionAlgorithmIdentifier'},
                                                                                    {'name': 'encryptedKey',
                                                                                     'type': 'EncryptedKey'}],
                                                                        'type': 'SEQUENCE'},
                                              'MessageAuthenticationCode': {'type': 'OCTET '
                                                                                    'STRING'},
                                              'MessageAuthenticationCodeAlgorithm': {'type': 'AlgorithmIdentifier'},
                                              'MessageDigest': {'type': 'OCTET '
                                                                        'STRING'},
                                              'OriginatorIdentifierOrKey': {'members': [{'name': 'issuerAndSerialNumber',
                                                                                         'type': 'IssuerAndSerialNumber'},
                                                                                        {'name': 'subjectKeyIdentifier',
                                                                                         'tag': {'number': 0},
                                                                                         'type': 'SubjectKeyIdentifier'},
                                                                                        {'name': 'originatorKey',
                                                                                         'tag': {'number': 1},
                                                                                         'type': 'OriginatorPublicKey'}],
                                                                            'type': 'CHOICE'},
                                              'OriginatorInfo': {'members': [{'name': 'certs',
                                                                              'optional': True,
                                                                              'tag': {'kind': 'IMPLICIT',
                                                                                      'number': 0},
                                                                              'type': 'CertificateSet'},
                                                                             {'name': 'crls',
                                                                              'optional': True,
                                                                              'tag': {'kind': 'IMPLICIT',
                                                                                      'number': 1},
                                                                              'type': 'RevocationInfoChoices'}],
                                                                 'type': 'SEQUENCE'},
                                              'OriginatorPublicKey': {'members': [{'name': 'algorithm',
                                                                                   'type': 'AlgorithmIdentifier'},
                                                                                  {'name': 'publicKey',
                                                                                   'type': 'BIT '
                                                                                           'STRING'}],
                                                                      'type': 'SEQUENCE'},
                                              'OtherCertificateFormat': {'members': [{'name': 'otherCertFormat',
                                                                                      'type': 'OBJECT '
                                                                                              'IDENTIFIER'},
                                                                                     {'choices': {},
                                                                                      'name': 'otherCert',
                                                                                      'type': 'ANY '
                                                                                              'DEFINED '
                                                                                              'BY',
                                                                                      'value': 'otherCertFormat'}],
                                                                         'type': 'SEQUENCE'},
                                              'OtherKeyAttribute': {'members': [{'name': 'keyAttrId',
                                                                                 'type': 'OBJECT '
                                                                                         'IDENTIFIER'},
                                                                                {'choices': {},
                                                                                 'name': 'keyAttr',
                                                                                 'optional': True,
                                                                                 'type': 'ANY '
                                                                                         'DEFINED '
                                                                                         'BY',
                                                                                 'value': 'keyAttrId'}],
                                                                    'type': 'SEQUENCE'},
                                              'OtherRecipientInfo': {'members': [{'name': 'oriType',
                                                                                  'type': 'OBJECT '
                                                                                          'IDENTIFIER'},
                                                                                 {'choices': {},
                                                                                  'name': 'oriValue',
                                                                                  'type': 'ANY '
                                                                                          'DEFINED '
                                                                                          'BY',
                                                                                  'value': 'oriType'}],
                                                                     'type': 'SEQUENCE'},
                                              'OtherRevocationInfoFormat': {'members': [{'name': 'otherRevInfoFormat',
                                                                                         'type': 'OBJECT '
                                                                                                 'IDENTIFIER'},
                                                                                        {'choices': {},
                                                                                         'name': 'otherRevInfo',
                                                                                         'type': 'ANY '
                                                                                                 'DEFINED '
                                                                                                 'BY',
                                                                                         'value': 'otherRevInfoFormat'}],
                                                                            'type': 'SEQUENCE'},
                                              'PasswordRecipientInfo': {'members': [{'name': 'version',
                                                                                     'type': 'CMSVersion'},
                                                                                    {'name': 'keyDerivationAlgorithm',
                                                                                     'optional': True,
                                                                                     'tag': {'number': 0},
                                                                                     'type': 'KeyDerivationAlgorithmIdentifier'},
                                                                                    {'name': 'keyEncryptionAlgorithm',
                                                                                     'type': 'KeyEncryptionAlgorithmIdentifier'},
                                                                                    {'name': 'encryptedKey',
                                                                                     'type': 'EncryptedKey'}],
                                                                        'type': 'SEQUENCE'},
                                              'RecipientEncryptedKey': {'members': [{'name': 'rid',
                                                                                     'type': 'KeyAgreeRecipientIdentifier'},
                                                                                    {'name': 'encryptedKey',
                                                                                     'type': 'EncryptedKey'}],
                                                                        'type': 'SEQUENCE'},
                                              'RecipientEncryptedKeys': {'element': {'type': 'RecipientEncryptedKey'},
                                                                         'type': 'SEQUENCE '
                                                                                 'OF'},
                                              'RecipientIdentifier': {'members': [{'name': 'issuerAndSerialNumber',
                                                                                   'type': 'IssuerAndSerialNumber'},
                                                                                  {'name': 'subjectKeyIdentifier',
                                                                                   'tag': {'number': 0},
                                                                                   'type': 'SubjectKeyIdentifier'}],
                                                                      'type': 'CHOICE'},
                                              'RecipientInfo': {'members': [{'name': 'ktri',
                                                                             'type': 'KeyTransRecipientInfo'},
                                                                            {'name': 'kari',
                                                                             'tag': {'number': 1},
                                                                             'type': 'KeyAgreeRecipientInfo'},
                                                                            {'name': 'kekri',
                                                                             'tag': {'number': 2},
                                                                             'type': 'KEKRecipientInfo'},
                                                                            {'name': 'pwri',
                                                                             'tag': {'number': 3},
                                                                             'type': 'PasswordRecipientInfo'},
                                                                            {'name': 'ori',
                                                                             'tag': {'number': 4},
                                                                             'type': 'OtherRecipientInfo'}],
                                                                'type': 'CHOICE'},
                                              'RecipientInfos': {'element': {'type': 'RecipientInfo'},
                                                                 'size': [(1,
                                                                           'MAX')],
                                                                 'type': 'SET '
                                                                         'OF'},
                                              'RecipientKeyIdentifier': {'members': [{'name': 'subjectKeyIdentifier',
                                                                                      'type': 'SubjectKeyIdentifier'},
                                                                                     {'name': 'date',
                                                                                      'optional': True,
                                                                                      'type': 'GeneralizedTime'},
                                                                                     {'name': 'other',
                                                                                      'optional': True,
                                                                                      'type': 'OtherKeyAttribute'}],
                                                                         'type': 'SEQUENCE'},
                                              'RevocationInfoChoice': {'members': [{'name': 'crl',
                                                                                    'type': 'CertificateList'},
                                                                                   {'name': 'other',
                                                                                    'tag': {'kind': 'IMPLICIT',
                                                                                            'number': 1},
                                                                                    'type': 'OtherRevocationInfoFormat'}],
                                                                       'type': 'CHOICE'},
                                              'RevocationInfoChoices': {'element': {'type': 'RevocationInfoChoice'},
                                                                        'type': 'SET '
                                                                                'OF'},
                                              'Signature': {'type': 'BIT '
                                                                    'STRING'},
                                              'SignatureAlgorithmIdentifier': {'type': 'AlgorithmIdentifier'},
                                              'SignatureValue': {'type': 'OCTET '
                                                                         'STRING'},
                                              'SignedAttributes': {'element': {'type': 'Attribute'},
                                                                   'size': [(1,
                                                                             'MAX')],
                                                                   'type': 'SET '
                                                                           'OF'},
                                              'SignedData': {'members': [{'name': 'version',
                                                                          'type': 'CMSVersion'},
                                                                         {'name': 'digestAlgorithms',
                                                                          'type': 'DigestAlgorithmIdentifiers'},
                                                                         {'name': 'encapContentInfo',
                                                                          'type': 'EncapsulatedContentInfo'},
                                                                         {'name': 'certificates',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 0},
                                                                          'type': 'CertificateSet'},
                                                                         {'name': 'crls',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 1},
                                                                          'type': 'RevocationInfoChoices'},
                                                                         {'name': 'signerInfos',
                                                                          'type': 'SignerInfos'}],
                                                             'type': 'SEQUENCE'},
                                              'SignerIdentifier': {'members': [{'name': 'issuerAndSerialNumber',
                                                                                'type': 'IssuerAndSerialNumber'},
                                                                               {'name': 'subjectKeyIdentifier',
                                                                                'tag': {'number': 0},
                                                                                'type': 'SubjectKeyIdentifier'}],
                                                                   'type': 'CHOICE'},
                                              'SignerInfo': {'members': [{'name': 'version',
                                                                          'type': 'CMSVersion'},
                                                                         {'name': 'sid',
                                                                          'type': 'SignerIdentifier'},
                                                                         {'name': 'digestAlgorithm',
                                                                          'type': 'DigestAlgorithmIdentifier'},
                                                                         {'name': 'signedAttrs',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 0},
                                                                          'type': 'SignedAttributes'},
                                                                         {'name': 'signatureAlgorithm',
                                                                          'type': 'SignatureAlgorithmIdentifier'},
                                                                         {'name': 'signature',
                                                                          'type': 'SignatureValue'},
                                                                         {'name': 'unsignedAttrs',
                                                                          'optional': True,
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 1},
                                                                          'type': 'UnsignedAttributes'}],
                                                             'type': 'SEQUENCE'},
                                              'SignerInfos': {'element': {'type': 'SignerInfo'},
                                                              'type': 'SET OF'},
                                              'SigningTime': {'type': 'Time'},
                                              'SubjectKeyIdentifier': {'type': 'OCTET '
                                                                               'STRING'},
                                              'Time': {'members': [{'name': 'utcTime',
                                                                    'type': 'UTCTime'},
                                                                   {'name': 'generalTime',
                                                                    'type': 'GeneralizedTime'}],
                                                       'type': 'CHOICE'},
                                              'UnauthAttributes': {'element': {'type': 'Attribute'},
                                                                   'size': [(1,
                                                                             'MAX')],
                                                                   'type': 'SET '
                                                                           'OF'},
                                              'UnprotectedAttributes': {'element': {'type': 'Attribute'},
                                                                        'size': [(1,
                                                                                  'MAX')],
                                                                        'type': 'SET '
                                                                                'OF'},
                                              'UnsignedAttributes': {'element': {'type': 'Attribute'},
                                                                     'size': [(1,
                                                                               'MAX')],
                                                                     'type': 'SET '
                                                                             'OF'},
                                              'UserKeyingMaterial': {'type': 'OCTET '
                                                                             'STRING'}},
                                    'values': {'id-contentType': {'type': 'OBJECT '
                                                                          'IDENTIFIER',
                                                                  'value': [('iso',
                                                                             1),
                                                                            ('member-body',
                                                                             2),
                                                                            ('us',
                                                                             840),
                                                                            ('rsadsi',
                                                                             113549),
                                                                            ('pkcs',
                                                                             1),
                                                                            ('pkcs9',
                                                                             9),
                                                                            3]},
                                               'id-countersignature': {'type': 'OBJECT '
                                                                               'IDENTIFIER',
                                                                       'value': [('iso',
                                                                                  1),
                                                                                 ('member-body',
                                                                                  2),
                                                                                 ('us',
                                                                                  840),
                                                                                 ('rsadsi',
                                                                                  113549),
                                                                                 ('pkcs',
                                                                                  1),
                                                                                 ('pkcs9',
                                                                                  9),
                                                                                 6]},
                                               'id-ct-authData': {'type': 'OBJECT '
                                                                          'IDENTIFIER',
                                                                  'value': [('iso',
                                                                             1),
                                                                            ('member-body',
                                                                             2),
                                                                            ('us',
                                                                             840),
                                                                            ('rsadsi',
                                                                             113549),
                                                                            ('pkcs',
                                                                             1),
                                                                            ('pkcs-9',
                                                                             9),
                                                                            ('smime',
                                                                             16),
                                                                            ('ct',
                                                                             1),
                                                                            2]},
                                               'id-ct-contentInfo': {'type': 'OBJECT '
                                                                             'IDENTIFIER',
                                                                     'value': [('iso',
                                                                                1),
                                                                               ('member-body',
                                                                                2),
                                                                               ('us',
                                                                                840),
                                                                               ('rsadsi',
                                                                                113549),
                                                                               ('pkcs',
                                                                                1),
                                                                               ('pkcs9',
                                                                                9),
                                                                               ('smime',
                                                                                16),
                                                                               ('ct',
                                                                                1),
                                                                               6]},
                                               'id-data': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': [('iso', 1),
                                                                     ('member-body',
                                                                      2),
                                                                     ('us',
                                                                      840),
                                                                     ('rsadsi',
                                                                      113549),
                                                                     ('pkcs',
                                                                      1),
                                                                     ('pkcs7',
                                                                      7),
                                                                     1]},
                                               'id-digestedData': {'type': 'OBJECT '
                                                                           'IDENTIFIER',
                                                                   'value': [('iso',
                                                                              1),
                                                                             ('member-body',
                                                                              2),
                                                                             ('us',
                                                                              840),
                                                                             ('rsadsi',
                                                                              113549),
                                                                             ('pkcs',
                                                                              1),
                                                                             ('pkcs7',
                                                                              7),
                                                                             5]},
                                               'id-encryptedData': {'type': 'OBJECT '
                                                                            'IDENTIFIER',
                                                                    'value': [('iso',
                                                                               1),
                                                                              ('member-body',
                                                                               2),
                                                                              ('us',
                                                                               840),
                                                                              ('rsadsi',
                                                                               113549),
                                                                              ('pkcs',
                                                                               1),
                                                                              ('pkcs7',
                                                                               7),
                                                                              6]},
                                               'id-envelopedData': {'type': 'OBJECT '
                                                                            'IDENTIFIER',
                                                                    'value': [('iso',
                                                                               1),
                                                                              ('member-body',
                                                                               2),
                                                                              ('us',
                                                                               840),
                                                                              ('rsadsi',
                                                                               113549),
                                                                              ('pkcs',
                                                                               1),
                                                                              ('pkcs7',
                                                                               7),
                                                                              3]},
                                               'id-messageDigest': {'type': 'OBJECT '
                                                                            'IDENTIFIER',
                                                                    'value': [('iso',
                                                                               1),
                                                                              ('member-body',
                                                                               2),
                                                                              ('us',
                                                                               840),
                                                                              ('rsadsi',
                                                                               113549),
                                                                              ('pkcs',
                                                                               1),
                                                                              ('pkcs9',
                                                                               9),
                                                                              4]},
                                               'id-signedData': {'type': 'OBJECT '
                                                                         'IDENTIFIER',
                                                                 'value': [('iso',
                                                                            1),
                                                                           ('member-body',
                                                                            2),
                                                                           ('us',
                                                                            840),
                                                                           ('rsadsi',
                                                                            113549),
                                                                           ('pkcs',
                                                                            1),
                                                                           ('pkcs7',
                                                                            7),
                                                                           2]},
                                               'id-signingTime': {'type': 'OBJECT '
                                                                          'IDENTIFIER',
                                                                  'value': [('iso',
                                                                             1),
                                                                            ('member-body',
                                                                             2),
                                                                            ('us',
                                                                             840),
                                                                            ('rsadsi',
                                                                             113549),
                                                                            ('pkcs',
                                                                             1),
                                                                            ('pkcs9',
                                                                             9),
                                                                            5]}}}}


"""Distinguished Encoding Rules (DER) codec.

"""

from . import DecodeTagError
from . import ber
from . import restricted_utc_time_to_datetime
from . import restricted_utc_time_from_datetime
from . import restricted_generalized_time_to_datetime
from . import restricted_generalized_time_from_datetime
from .compiler import clean_bit_string_value
from .ber import Class
from .ber import Encoding
from .ber import Tag
from .ber import encode_length_definite
from .ber import decode_length_definite
from .ber import encode_signed_integer
from .ber import decode_signed_integer
from .ber import encode_tag
from .ber import Boolean
from .ber import Null
from .ber import ObjectIdentifier
from .ber import Enumerated
from .ber import Sequence
from .ber import Set
from .ber import Choice
from .ber import Any
from .ber import AnyDefinedBy
from .ber import Recursive
from .ber import decode_length


class Type(object):

    def __init__(self, name, type_name, number, flags=0):
        self.name = name
        self.type_name = type_name

        if number is None:
            self.tag = None
        else:
            self.tag = encode_tag(number, flags)

        self.optional = False
        self.default = None

    def set_tag(self, number, flags):
        if not Class.APPLICATION & flags:
            flags |= Class.CONTEXT_SPECIFIC

        self.tag = encode_tag(number, flags)

    def set_size_range(self, minimum, maximum, has_extension_marker):
        pass

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)

        if data[offset:end_offset] != self.tag:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

        return end_offset

    def is_default(self, value):
        return value == self.default


class StringType(Type):

    TAG = None
    ENCODING = None

    def __init__(self, name):
        super(StringType, self).__init__(name,
                                         self.__class__.__name__,
                                         self.TAG)

    def encode(self, data, encoded):
        data = data.encode(self.ENCODING)
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode(self.ENCODING), end_offset

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.name)


class ArrayType(Type):

    def __init__(self, name, tag_name, tag, element_type):
        super(ArrayType, self).__init__(name,
                                        tag_name,
                                        tag,
                                        Encoding.CONSTRUCTED)
        self.element_type = element_type

    def set_tag(self, number, flags):
        super(ArrayType, self).set_tag(number,
                                       flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_elements = bytearray()

        for entry in data:
            self.element_type.encode(entry, encoded_elements)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_elements)))
        encoded.extend(encoded_elements)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        decoded = []
        start_offset = offset

        while (offset - start_offset) < length:
            decoded_element, offset = self.element_type.decode(data, offset)
            decoded.append(decoded_element)

        return decoded, offset

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name,
                                      'INTEGER',
                                      Tag.INTEGER)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_signed_integer(data))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return decode_signed_integer(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL', Tag.REAL)

    def encode(self, data, encoded):
        raise NotImplementedError('REAL is not yet implemented.')

    def decode(self, data, offset):
        raise NotImplementedError('REAL is not yet implemented.')

    def __repr__(self):
        return 'Real({})'.format(self.name)


class BitString(Type):

    def __init__(self, name, has_named_bits):
        super(BitString, self).__init__(name,
                                        'BIT STRING',
                                        Tag.BIT_STRING)
        self.has_named_bits = has_named_bits

    def is_default(self, value):
        if self.default is None:
            return False

        clean_value = clean_bit_string_value(value,
                                             self.has_named_bits)
        clean_default = clean_bit_string_value(self.default,
                                               self.has_named_bits)

        return clean_value == clean_default

    def encode(self, data, encoded):
        number_of_bytes, number_of_rest_bits = divmod(data[1], 8)
        data = bytearray(data[0])

        if number_of_rest_bits == 0:
            data = data[:number_of_bytes]
            number_of_unused_bits = 0
        else:
            last_byte = data[number_of_bytes]
            last_byte &= ((0xff >> number_of_rest_bits) ^ 0xff)
            data = data[:number_of_bytes]
            data.append(last_byte)
            number_of_unused_bits = (8 - number_of_rest_bits)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data) + 1))
        encoded.append(number_of_unused_bits)
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        number_of_bits = 8 * (length - 1) - data[offset]
        offset += 1

        return (bytes(data[offset:end_offset]), number_of_bits), end_offset

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name,
                                          'OCTET STRING',
                                          Tag.OCTET_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return bytes(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         'SEQUENCE OF',
                                         Tag.SEQUENCE,
                                         element_type)


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    'SET OF',
                                    Tag.SET,
                                    element_type)


class UTF8String(StringType):

    TAG = Tag.UTF8_STRING
    ENCODING = 'utf-8'


class NumericString(StringType):

    TAG = Tag.NUMERIC_STRING
    ENCODING = 'ascii'


class PrintableString(StringType):

    TAG = Tag.PRINTABLE_STRING
    ENCODING = 'ascii'


class IA5String(StringType):

    TAG = Tag.IA5_STRING
    ENCODING = 'ascii'


class VisibleString(StringType):

    TAG = Tag.VISIBLE_STRING
    ENCODING = 'ascii'


class GeneralString(StringType):

    TAG = Tag.GENERAL_STRING
    ENCODING = 'latin-1'


class BMPString(StringType):

    TAG = Tag.BMP_STRING
    ENCODING = 'utf-16-be'


class UniversalString(StringType):

    TAG = Tag.UNIVERSAL_STRING
    ENCODING = 'utf-32-be'


class GraphicString(StringType):

    TAG = Tag.GRAPHIC_STRING
    ENCODING = 'latin-1'


class TeletexString(StringType):

    TAG = Tag.T61_STRING
    ENCODING = 'iso-8859-1'


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name,
                                      'UTCTime',
                                      Tag.UTC_TIME)

    def encode(self, data, encoded):
        data = restricted_utc_time_from_datetime(data).encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')

        return restricted_utc_time_to_datetime(decoded), end_offset

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name,
                                              'GeneralizedTime',
                                              Tag.GENERALIZED_TIME)

    def encode(self, data, encoded):
        data = restricted_generalized_time_from_datetime(data).encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')

        return restricted_generalized_time_to_datetime(decoded), end_offset

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class Compiler(ber.Compiler):

    def compile_implicit_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            has_named_bits = ('named-bits' in type_descriptor)
            compiled = BitString(name, has_named_bits)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            choices = {}

            for key, value in type_descriptor['choices'].items():
                choices[key] = self.compile_type(key,
                                                 value,
                                                 module_name)

            compiled = AnyDefinedBy(name,
                                    type_descriptor['value'],
                                    choices)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        return compiled


def compile_dict(specification):
    return Compiler(specification).process()
